package dev.fyloz.trial.colorrecipesexplorer.model

import com.fasterxml.jackson.annotation.JsonIgnore
import com.fasterxml.jackson.annotation.JsonProperty
import dev.fyloz.trial.colorrecipesexplorer.model.validation.NullOrNotBlank
import org.hibernate.annotations.Fetch
import org.hibernate.annotations.FetchMode
import org.springframework.security.core.GrantedAuthority
import org.springframework.security.core.authority.SimpleGrantedAuthority
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder
import org.springframework.security.crypto.password.PasswordEncoder
import java.time.LocalDateTime
import java.util.*
import javax.persistence.*
import javax.validation.constraints.NotBlank
import javax.validation.constraints.NotNull
import javax.validation.constraints.Size


private const val EMPLOYEE_ID_NULL_MESSAGE = "Un numéro d'employé est requis"
private const val EMPLOYEE_LAST_NAME_EMPTY_MESSAGE = "Un nom  est requis"
private const val EMPLOYEE_FIRST_NAME_EMPTY_MESSAGE = "Un prénom est requis"
private const val EMPLOYEE_PASSWORD_EMPTY_MESSAGE = "Un mot de passe est requis"
private const val EMPLOYEE_PASSWORD_TOO_SHORT_MESSAGE = "Le mot de passe doit contenir au moins 8 caractères"

@Entity
@Table(name = "employee")
data class Employee(
    @Id
    override val id: Long,

    @Column(name = "first_name")
    val firstName: String = "",

    @Column(name = "last_name")
    val lastName: String = "",

    @JsonIgnore
    val password: String = "",

    @JsonIgnore
    @Column(name = "default_group_user")
    val isDefaultGroupUser: Boolean = false,

    @JsonIgnore
    @Column(name = "system_user")
    val isSystemUser: Boolean = false,

    @ManyToOne
    @JoinColumn(name = "group_id")
    @Fetch(FetchMode.SELECT)
    var group: EmployeeGroup? = null,

    @Enumerated(EnumType.STRING)
    @ElementCollection(fetch = FetchType.EAGER)
    @CollectionTable(name = "employee_permission", joinColumns = [JoinColumn(name = "employee_id")])
    @Column(name = "permission")
    @Fetch(FetchMode.SUBSELECT)
    @get:JsonProperty("explicitPermissions")
    val permissions: MutableSet<EmployeePermission> = mutableSetOf(),

    @Column(name = "last_login_time")
    var lastLoginTime: LocalDateTime? = null
) : Model {
    @JsonProperty("permissions")
    fun getFlattenedPermissions(): Iterable<EmployeePermission> = getPermissions()

    override fun equals(other: Any?): Boolean =
        other is Employee && id == other.id && firstName == other.firstName && lastName == other.lastName

    override fun hashCode(): Int = Objects.hash(id, firstName, lastName)
}

/** DTO for creating employees. Allows a [password] a [groupId]. */
open class EmployeeSaveDto(
    @field:NotNull(message = EMPLOYEE_ID_NULL_MESSAGE)
    val id: Long,

    @field:NotBlank(message = EMPLOYEE_FIRST_NAME_EMPTY_MESSAGE)
    val firstName: String,

    @field:NotBlank(message = EMPLOYEE_LAST_NAME_EMPTY_MESSAGE)
    val lastName: String,

    @field:NotBlank(message = EMPLOYEE_PASSWORD_EMPTY_MESSAGE)
    @field:Size(min = 8, message = EMPLOYEE_PASSWORD_TOO_SHORT_MESSAGE)
    val password: String,

    @field:ManyToOne
    @Fetch(FetchMode.SELECT)
    var groupId: Long? = null,

    @Enumerated(EnumType.STRING)
    val permissions: MutableSet<EmployeePermission> = mutableSetOf()
) : EntityDto<Employee> {
    override fun toEntity(): Employee =
        Employee(
            id,
            firstName,
            lastName,
            "",
            isDefaultGroupUser = false,
            isSystemUser = false,
            group = null,
            permissions = permissions
        )
}

open class EmployeeUpdateDto(
    @field:NotNull(message = EMPLOYEE_ID_NULL_MESSAGE)
    val id: Long,

    @field:NullOrNotBlank(message = EMPLOYEE_FIRST_NAME_EMPTY_MESSAGE)
    val firstName: String?,

    @field:NullOrNotBlank(message = EMPLOYEE_LAST_NAME_EMPTY_MESSAGE)
    val lastName: String?,

    @Enumerated(EnumType.STRING)
    val permissions: Set<EmployeePermission>?
) : EntityDto<Employee>


private const val GROUP_ID_NULL_MESSAGE = "Un identifiant est requis"
private const val GROUP_NAME_NULL_MESSAGE = "Un nom est requis"
private const val GROUP_PERMISSIONS_EMPTY_MESSAGE = "Au moins une permission est requise"

@Entity
@Table(name = "employee_group")
data class EmployeeGroup(
    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    override var id: Long? = null,

    @Column(unique = true)
    override val name: String = "",

    @Enumerated(EnumType.STRING)
    @ElementCollection(fetch = FetchType.EAGER)
    @CollectionTable(name = "group_permission", joinColumns = [JoinColumn(name = "group_id")])
    @Column(name = "permission")
    @Fetch(FetchMode.SUBSELECT)
    val permissions: MutableSet<EmployeePermission> = mutableSetOf(),

    @OneToMany(mappedBy = "group")
    @JsonIgnore
    val employees: MutableSet<Employee> = mutableSetOf()
) : NamedModel {
    @JsonProperty("employeeCount")
    fun getEmployeeCount() = employees.size - 1 // Remove the default employee

    override fun equals(other: Any?): Boolean = other is EmployeeGroup && id == other.id && name == other.name
    override fun hashCode(): Int = Objects.hash(id, name)
}

open class EmployeeGroupSaveDto(
    @field:NotBlank(message = GROUP_NAME_NULL_MESSAGE)
    @field:Size(min = 3)
    val name: String,

    @field:Size(min = 1, message = GROUP_PERMISSIONS_EMPTY_MESSAGE)
    val permissions: MutableSet<EmployeePermission>
) : EntityDto<EmployeeGroup> {
    override fun toEntity(): EmployeeGroup =
        EmployeeGroup(null, name, permissions)
}

open class EmployeeGroupUpdateDto(
    @field:NotNull(message = GROUP_ID_NULL_MESSAGE)
    val id: Long,

    @field:NotBlank(message = GROUP_NAME_NULL_MESSAGE)
    @field:Size(min = 3)
    val name: String,

    @field:Size(min = 1, message = GROUP_PERMISSIONS_EMPTY_MESSAGE)
    val permissions: MutableSet<EmployeePermission>
) : EntityDto<EmployeeGroup> {
    override fun toEntity(): EmployeeGroup =
        EmployeeGroup(id, name, permissions)
}


data class EmployeeLoginRequest(val id: Long, val password: String)


enum class EmployeePermission(val impliedPermissions: List<EmployeePermission> = listOf()) {
    // View
    VIEW_MATERIAL,
    VIEW_MATERIAL_TYPE,
    VIEW_COMPANY,
    VIEW_RECIPE,
    VIEW(
        listOf(
            VIEW_MATERIAL,
            VIEW_MATERIAL_TYPE,
            VIEW_COMPANY,
            VIEW_RECIPE
        )
    ),
    VIEW_EMPLOYEE,
    VIEW_EMPLOYEE_GROUP,

    // Edit
    EDIT_MATERIAL(listOf(VIEW_MATERIAL)),
    EDIT_MATERIAL_TYPE(listOf(VIEW_MATERIAL_TYPE)),
    EDIT_COMPANY(listOf(VIEW_COMPANY)),
    EDIT_RECIPE(listOf(VIEW_RECIPE)),
    EDIT(
        listOf(
            EDIT_MATERIAL,
            EDIT_MATERIAL_TYPE,
            EDIT_COMPANY,
            EDIT_RECIPE,
            VIEW
        )
    ),
    EDIT_EMPLOYEE(listOf(VIEW_EMPLOYEE)),
    EDIT_EMPLOYEE_PASSWORD(listOf(EDIT_EMPLOYEE)),
    EDIT_EMPLOYEE_GROUP(listOf(VIEW_EMPLOYEE_GROUP)),

    // Remove
    REMOVE_MATERIAL(listOf(EDIT_MATERIAL)),
    REMOVE_MATERIAL_TYPE(listOf(EDIT_MATERIAL_TYPE)),
    REMOVE_COMPANY(listOf(EDIT_COMPANY)),
    REMOVE_RECIPE(listOf(EDIT_RECIPE)),
    REMOVE(
        listOf(
            REMOVE_MATERIAL,
            REMOVE_MATERIAL_TYPE,
            REMOVE_COMPANY,
            REMOVE_RECIPE,
            EDIT
        )
    ),
    REMOVE_EMPLOYEE(listOf(EDIT_EMPLOYEE)),
    REMOVE_EMPLOYEE_GROUP(listOf(EDIT_EMPLOYEE_GROUP)),

    // Others
    SET_BROWSER_DEFAULT_GROUP(
        listOf(
            VIEW_EMPLOYEE_GROUP
        )
    ),

    ADMIN(
        listOf(
            REMOVE,
            SET_BROWSER_DEFAULT_GROUP,

            // Admin only permissions
            REMOVE_EMPLOYEE,
            EDIT_EMPLOYEE_PASSWORD,
            REMOVE_EMPLOYEE_GROUP,
        )
    );

    operator fun contains(permission: EmployeePermission): Boolean {
        return permission == this || impliedPermissions.any { permission in it }
    }
}

/** Gets [GrantedAuthority]s of the given [Employee]. */
fun Employee.getAuthorities(): MutableCollection<GrantedAuthority> {
    return getPermissions().map { it.toAuthority() }.toMutableSet()
}

/** Gets [EmployeePermission]s of the given [Employee]. */
fun Employee.getPermissions(): Iterable<EmployeePermission> {
    val grantedPermissions: MutableSet<EmployeePermission> = mutableSetOf()
    if (group != null) grantedPermissions.addAll(group!!.permissions.flatMap { it.flat() })
    grantedPermissions.addAll(permissions.flatMap { it.flat() })
    return grantedPermissions
}

private fun EmployeePermission.flat(): Iterable<EmployeePermission> {
    return mutableSetOf(this).apply {
        impliedPermissions.forEach {
            addAll(it.flat())
        }
    }
}

/** Converts the given [EmployeePermission] to a [GrantedAuthority]. */
private fun EmployeePermission.toAuthority(): GrantedAuthority {
    return SimpleGrantedAuthority(name)
}


// ==== DSL ====
fun employee(
    passwordEncoder: PasswordEncoder = BCryptPasswordEncoder(),
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    password: String = passwordEncoder.encode("password"),
    isDefaultGroupUser: Boolean = false,
    isSystemUser: Boolean = false,
    group: EmployeeGroup? = null,
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    lastLoginTime: LocalDateTime? = null,
    op: Employee.() -> Unit = {}
) = Employee(
    id,
    firstName,
    lastName,
    password,
    isDefaultGroupUser,
    isSystemUser,
    group,
    permissions,
    lastLoginTime
).apply(op)

fun employeeSaveDto(
    passwordEncoder: PasswordEncoder = BCryptPasswordEncoder(),
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    password: String = passwordEncoder.encode("password"),
    groupId: Long? = null,
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    op: EmployeeSaveDto.() -> Unit = {}
) = EmployeeSaveDto(id, firstName, lastName, password, groupId, permissions).apply(op)

fun employeeUpdateDto(
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    op: EmployeeUpdateDto.() -> Unit = {}
) = EmployeeUpdateDto(id, firstName, lastName, permissions).apply(op)

fun employeeGroup(
    id: Long? = null,
    name: String = "name",
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    employees: MutableSet<Employee> = mutableSetOf(),
    op: EmployeeGroup.() -> Unit = {}
) = EmployeeGroup(id, name, permissions, employees).apply(op)

fun employeeGroupSaveDto(
    name: String = "name",
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    op: EmployeeGroupSaveDto.() -> Unit = {}
) = EmployeeGroupSaveDto(name, permissions).apply(op)

fun employeeGroupUpdateDto(
    id: Long = 0L,
    name: String = "name",
    permissions: MutableSet<EmployeePermission> = mutableSetOf(),
    op: EmployeeGroupUpdateDto.() -> Unit = {}
) = EmployeeGroupUpdateDto(id, name, permissions).apply(op)
